'''
    Python wrapper around curve25519 by mehdi sotoodeh.
    The files under py25519/c  are from mehdi sotoodeh and are copied
    unchanged from https://github.com/msotoodeh/curve25519.

    The files under py25519/c are licensed under the MIT LICENSE (see
    license.txt file under py25519/c).

    All remaining files in this package are licensed under the GNU
    General Public License version 3 or (at your option) any later version.

    See the file LICENSE-GPLv3.txt for details of the GNU General Public
    License version 3.

    Copyright (C) 2016  Sundar Nagarajan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
from cffi_utils.utils2to3 import inputToBytes
from cffi_utils.sowrapper import get_lib_ffi_resource
import os


c_hdr = '''
void ed25519_CreateKeyPair(
    unsigned char *pubKey,
    unsigned char *privKey,
    const void *blinding,
    const unsigned char *sk);
void ed25519_SignMessage(
    unsigned char *signature,
    const unsigned char *privKey,
    const void *blinding,
    const unsigned char *msg,
    size_t msg_size);
int ed25519_VerifySignature(
    const unsigned char *signature,
    const unsigned char *publicKey,
    const unsigned char *msg,
    size_t msg_size);
void curve25519_dh_CalculatePublicKey(
    unsigned char *pk,
    unsigned char *sk);
void curve25519_dh_CalculatePublicKey_fast(
    unsigned char *pk,
    unsigned char *sk);
void curve25519_dh_CreateSharedKey(
    unsigned char *shared,
    const unsigned char *pk,
    unsigned char *sk);
'''

module_name = 'py25519'
libpath = 'libcurve25519.so'

(ffi, lib) = get_lib_ffi_resource(
    module_name=module_name, libpath=libpath, c_hdr=c_hdr)

SECRETKEY_SIZE = 32
VERIFYINGKEY_SIZE = 32
SIGNINGKEY_SIZE = 64
SIGNATURE_SIZE = 64
PUBLICKEY_SIZE = 32
SHAREDKEY_SIZE = 32
C_VERIFYINGKEY = 'unsigned char[%d]' % (VERIFYINGKEY_SIZE,)
C_SIGNINGKEY = 'unsigned char[%d]' % (SIGNINGKEY_SIZE,)
C_SECRETKEY = 'unsigned char[%d]' % (SECRETKEY_SIZE,)
C_SIGNATURE = 'unsigned char[%d]' % (SIGNATURE_SIZE,)
C_PUBLICKEY = 'unsigned char[%d]' % (PUBLICKEY_SIZE,)
C_SHAREDKEY = 'unsigned char[%d]' % (SHAREDKEY_SIZE,)


class KeyTypeError(Exception):
    pass


class Key25519(object):
    def __init__(self, secretkey=None,
                 verifyingkey=None, pubkey=None):
        '''
        secretkey-->bytes[32]
        verifyingkey-->bytes[32]
        pubkey-->bytes[32]
        A PublicKey version of the object will have ONLY verifyingkey
        and pubkey and signingkey will be None
        A PrivateKey version of the object will have secretkey set
        and verifyingkey and pubkey are ignored and derived from secretkey

        Storing and reusing the secretkey is a way to recreate your key pair
        '''
        # PublicKey version
        if secretkey is None and (verifyingkey is not None and
                                  pubkey is not None):
            self.__secretkey = None
            self.__verifyingkey = ffi.new(C_VERIFYINGKEY, verifyingkey)
            self.__pubkey = ffi.new(C_PUBLICKEY, pubkey)
            return
        # PrivateKey version - recreation
        if secretkey is not None:
            self.__secretkey = ffi.new(C_SECRETKEY, secretkey)
        # PrivateKey version - new object
        else:
            self.__secretkey = ffi.new(C_SECRETKEY,
                                       os.urandom(SECRETKEY_SIZE))
        # PrivateKey recreation or new but not PublicKey
        self.__verifyingkey = ffi.new(C_VERIFYINGKEY)
        self.__signingkey = ffi.new(C_SIGNINGKEY)
        lib.ed25519_CreateKeyPair(
            self.__verifyingkey,
            self.__signingkey,
            ffi.NULL,
            self.__secretkey)
        self.__pubkey = ffi.new(C_PUBLICKEY)
        # Copy because curve25519_dh_CalculatePublicKey_fast changes secretkey
        tmp_secretkey = ffi.new(C_SECRETKEY, self.secretkey)
        lib.curve25519_dh_CalculatePublicKey_fast(
            self.__pubkey, tmp_secretkey)
        del tmp_secretkey
        self.selftest()

    def selftest(self):
        '''
        Returns-->boolean
        '''
        msg = 'ABCD'
        # If it is a PublicKey version, we need a PrivateKey to test against
        if self.signingkey is None:
            pvtkey = Key25519()
        else:
            pvtkey = self
        try:
            sig = pvtkey.sign(msg)
            if not self.verify(sig, msg):
                raise ValueError('Invalid values for key')
            pvtkey.get_ecdh_key(self.public_key())
        except:
            raise ValueError('Invalid values for key')

    @inputToBytes
    def sign(self, msg):
        '''
        msg-->str or bytes
        Returns-->bytes
        '''
        if self.__signingkey is None:
            raise KeyTypeError('Wrong key type for operation')
        sig = ffi.new(C_SIGNATURE)
        lib.ed25519_SignMessage(
            sig, self.__signingkey, ffi.NULL, msg, len(msg))
        return ffi.get_bytes(sig)

    @inputToBytes
    def verify(self, sig, msg):
        '''
        sig-->bytes
        msg-->str or bytes
        Returns-->boolean
        '''
        return lib.ed25519_VerifySignature(
            sig, self.__verifyingkey, msg, len(msg)) == 1

    def get_ecdh_key(self, other):
        '''
        other-->Key25519 instance
        '''
        if self.__secretkey is None:
            raise KeyTypeError('Wrong key type for operation')
        # Copy because curve25519_dh_CreateSharedKey changes secretkey
        tmp_secretkey = ffi.new(C_SECRETKEY, self.secretkey)
        ecdh_key = ffi.new(C_SHAREDKEY)
        lib.curve25519_dh_CreateSharedKey(
            ecdh_key, other.pubkey, tmp_secretkey)
        del tmp_secretkey
        return ffi.get_bytes(ecdh_key)

    def public_key(self):
        '''
        Returns a PublicKey version of this object
        '''
        return Key25519(verifyingkey=self.verifyingkey,
                        pubkey=self.pubkey)

    @property
    def secretkey(self):
        if self.__secretkey is None:
            return None
        return ffi.get_bytes(self.__secretkey)

    @property
    def signingkey(self):
        if self.__signingkey is None:
            return None
        return ffi.get_bytes(self.__signingkey)

    @property
    def verifyingkey(self):
        if self.__verifyingkey is None:
            return None
        return ffi.get_bytes(self.__verifyingkey)

    @property
    def pubkey(self):
        if self.__pubkey is None:
            return None
        return ffi.get_bytes(self.__pubkey)

    def serialize(self):
        if self.__signingkey is None:
            return self.pubkey
        else:
            return self.secretkey

    def __eq__(self, other):
        attrs = ['secretkey', 'signingkey',
                 'verifyingkey', 'pubkey']
        a = [getattr(self, x, None) for x in attrs]
        b = [getattr(other, x, None) for x in attrs]
        return a == b



from . import Key25519


if __name__ == '__main__':
    c = Key25519()
    print('secretkey: ', c.secretkey)
    print('signingkey: ', c.signingkey)
    print('verifyingkey: ', c.verifyingkey)

    msg = 'Hello world'
    badmsg = msg + '1'
    sig = c.sign(msg)
    badsig = c.sign(badmsg)
    print('good verify: ', c.verify(sig, msg))
    print('badmsg verify: ', c.verify(sig, badmsg))
    print('badsig verify: ', c.verify(badsig, msg))

    d = Key25519(secretkey=c.secretkey)
    print('----- d is a copy of c -----')
    print('d.verify: ', d.verify(sig, msg))
    print('c == d:', c == d)
    print('public(c) == public(d):', c.public_key() == d.public_key())
    print('c == public(d):', c == d.public_key())

    print('----- e is a new different key -----')
    e = Key25519()
    print('e.verify: ', e.verify(sig, msg))
    print('c == e:', c == e)
    print('public(c) == public(e):', c.public_key() == e.public_key())

    print('----- Test ECDH shared key -----')
    ecdh1 = c.get_ecdh_key(e.public_key())
    ecdh2 = e.get_ecdh_key(c.public_key())
    print('ecdh equal: ', ecdh1 == ecdh2)
    ecdh3 = e.get_ecdh_key(c.public_key())
    print('ecdh equal again: ', ecdh1 == ecdh3)



'''
    setupext.py: A python module with utility functions to ease creation
    of python setup (installation) scripts

    Copyright (C) 2016 Sundar Nagarajan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    For details of the GNU General Pulic License version 3, see the
    LICENSE-GPLv3.txt file that accompanied this program
'''
import sys
import os
import subprocess
import traceback
from distutils.command.build import build
from distutils.command.build_clib import build_clib
from distutils.command.build_ext import build_ext
from distutils.command.build_py import build_py
from distutils.command.build_scripts import build_scripts
from distutils.command.install import install
from distutils.command.install_data import install_data
from distutils.command.install_headers import install_headers
from distutils.command.install_lib import install_lib
from distutils.command.install_scripts import install_scripts


# Patch this variable from outside the module
config = {}
known_steps = [
    'build',
    'build_clib',
    'build_ext',
    'build_py',
    'build_scripts',
    'install_data',
    'install_lib',
    'install_headers',
    # Following do not work
    # 'install',
    # 'install_scripts',
]
trace_triggers = False

# setup config, so that user only needs to add cmdlist and optionally
# set show_output, show_err, ignore_err for different triggers
for step in known_steps:
    config[step] = {}
    for pre_post in ['pre', 'post']:
        config[step][pre_post] = {}
del step, pre_post


def get_dirtree(module_name, dirlist=[]):
    '''
    module_name-->str: must be name of a dir under current working dir
    dirlist-->list of str: must all be names of dirs under module_name
    '''
    ret = []
    curdir = os.getcwd()
    if not os.path.isdir(module_name):
        return ret
    os.chdir(module_name)
    try:
        for dirname in dirlist:
            if not os.path.isdir(dirname):
                continue
            for (d, ds, fs) in os.walk(dirname):
                for f in fs:
                    ret += [os.path.join(d, f)]
        return ret
    except:
        return ret
    finally:
        os.chdir(curdir)


def get_cmdclass():
    '''
    Returns-->dict: to be assigned to setuptools.setup
                    cmdclass param
    '''
    ret = {}
    cfg = __get_config()
    if cfg is None:
        return
    mapdict = {
        'build': CustomBuild,
        'build_clib': CustomBuildCLib,
        'build_ext': CustomBuildExt,
        'build_py': CustomBuildPy,
        'build_scripts': CustomBuildScripts,
        'install': CustomInstall,
        'install_data': CustomInstallData,
        'install_headers': CustomInstallHeaders,
        'install_lib': CustomInstallLib,
        'install_scripts': CustomInstallScripts,
    }
    for step in cfg.keys():
        if step in mapdict:
            ret[step] = mapdict[step]
    return ret


def run_in_order(l, show_output=True, show_err=True, ignore_err=False,
                 args=(), **kwargs):
    '''
    Processes each element of l in order:
        if it is a string: execute it as a shell command
        elif it is a callable, call it with *args, **kwargs

    l-->list: Each elem is either a string (shell command) or callable
              Any other type is ignored
    show_output-->boolean: Show stdout of shell commands
                           Does not affect callables
    show_err-->Boolean: Show stderr of shell commands
                        Does not affect callables
    ignore_err-->boolean: Continue after exception or shell command
                          wth return code != 0
    Returns-->Nothing

    if ignore_err == False, exceptions are re-raised, hence shown

    ------------------------------------------------------------------
    show_output show_err ignore_err   stdout stderr exception continue
                                                    trace
    ------------------------------------------------------------------
    True        True     False        SHOW   SHOW   SHOW      NO
    True        False    False        SHOW   HIDE   SHOW      NO
    False       True     False        HIDE   SHOW   SHOW      NO
    False       False    False        HIDE   HIDE   SHOW      NO

    True        True     True         SHOW   SHOW   SHOW      YES
    True        False    True         SHOW   HIDE   HIDE      YES
    False       True     True         HIDE   SHOW   SHOW      YES
    False       False    True         HIDE   HIDE   HIDE      YES
    ------------------------------------------------------------------
    ----------- DEFAULT -----------   SHOW   SHOW   SHOW      NO
    ------------------------------------------------------------------
    '''
    # Set defaults
    if show_output is None:
        show_output = True
    if show_err is None:
        show_err = True
    if ignore_err is None:
        ignore_err = False
    if args is None:
        args = ()

    for c in l:
        try:
            if isinstance(c, str):
                devnull = open(os.devnull, 'w')
                if not show_err:
                    stderr = devnull
                else:
                    stderr = None
                if not show_output:
                    stdout = devnull
                else:
                    stdout = None
                retcode = subprocess.call(
                    c, shell=True, stdout=stdout, stderr=stderr)
                if not ignore_err and retcode != 0:
                    break
            elif hasattr(c, '__call__'):
                c(*args, **kwargs)
        except:
            if not ignore_err:
                raise
            if show_err:
                sys.stderr.write(traceback.format_exc())


def run_triggers(step, pre_post, args=(), **kwargs):
    '''
    step, pre_post-->str: Looked up in CONFIG_VAR
                          step must be key in CONFIG_VAR
                          pre_post must be key in CONFIG_VAR[step]
    args-->tuple: Passed to run_in_order()
    '''
    if not isinstance(step, str) or not isinstance(pre_post, str):
        return
    if trace_triggers:
        msg = 'Trigger for %s: %s' % (step, pre_post)
        sys.stderr.write(msg.center(80, '-') + '\n')
    if args is None:
        args = ()
    if not isinstance(args, tuple):
        return
    cfg = __get_config()
    if cfg is None:
        return
    if step not in cfg:
        return
    if not isinstance(cfg[step], dict):
        return
    if pre_post not in cfg[step]:
        return
    d = cfg[step][pre_post]
    l = d.get('cmdlist', [])
    if not l:
        return

    show_output = d.get('show_output', None)
    show_err = d.get('show_err', None)
    ignore_err = d.get('ignore_err', None)

    run_in_order(
        l,
        show_output=show_output,
        show_err=show_err,
        ignore_err=ignore_err,
        args=args,
        pre_post=pre_post,
        kwargs=kwargs
    )


def __get_config():
    '''
    Returns-->dict or None
    '''
    if 'config' not in globals():
        return None
    cfg = globals().get('config', None)
    if cfg is None:
        return None

    if not isinstance(cfg, dict):
        return None
    return cfg


class CustomBuild(build):
    def run(self):
        run_triggers('build', 'pre', args=(self,))
        build.run(self)
        run_triggers('build', 'post', args=(self,))


class CustomBuildCLib(build_clib):
    def run(self):
        run_triggers('build_clib', 'pre', args=(self,))
        build_clib.run(self)
        run_triggers('build_clib', 'post', args=(self,))


class CustomBuildExt(build_ext):
    def run(self):
        run_triggers('build_ext', 'pre', args=(self,))
        build_ext.run(self)
        run_triggers('build_ext', 'post', args=(self,))


class CustomBuildPy(build_py):
    def run(self):
        run_triggers('build_py', 'pre', args=(self,))
        build_py.run(self)
        run_triggers('build_py', 'post', args=(self,))


class CustomBuildScripts(build_scripts):
    def run(self):
        run_triggers('build_scripts', 'pre', args=(self,))
        build_scripts.run(self)
        run_triggers('build_scripts', 'post', args=(self,))


class CustomInstall(install):
    def run(self):
        # run_triggers('install', 'pre', args=(self,))
        install.run(self)
        run_triggers('install', 'post', args=(self,))


class CustomInstallData(install_data):
    def run(self):
        run_triggers('install_data', 'pre', args=(self,))
        install_data.run(self)
        run_triggers('install_data', 'post', args=(self,))


class CustomInstallHeaders(install_headers):
    def run(self):
        run_triggers('install_headers', 'pre', args=(self,))
        install_headers.run(self)
        run_triggers('install_headers', 'post', args=(self,))


class CustomInstallLib(install_lib):
    def run(self):
        run_triggers('install_lib', 'pre', args=(self,))
        install_lib.run(self)
        run_triggers('install_lib', 'post', args=(self,))


class CustomInstallScripts(install_scripts):
    def run(self):
        run_triggers('install_scripts', 'pre', args=(self,))
        install_scripts.run(self)
        run_triggers('install_scripts', 'post', args=(self,))


if __name__ == '__main__':
    def test(*args, **kwargs):
        if len(args) < 2:
            raise ValueError('test: Insufficient arguments')
        if len(kwargs) == 0:
            sys.stderr.write('test: kwargs is empty\n')
        print('test: args: ' + str(args))

    l = ['uname -a', 'uname --nosuchoption', test, 'uname']
    run_in_order(l, ignore_err=True, args=(1, 2))



'''
    Python wrapper around curve25519 by mehdi sotoodeh.
    The files under py25519/c  are from mehdi sotoodeh and are copied
    unchanged from https://github.com/msotoodeh/curve25519.

    The files under py25519/c are licensed under the MIT LICENSE (see
    license.txt file under py25519/c).

    All remaining files in this package are licensed under the GNU
    General Public License version 3 or (at your option) any later version.

    See the file LICENSE-GPLv3.txt for details of the GNU General Public
    License version 3.

    Copyright (C) 2016  Sundar Nagarajan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
import sys
import os
from setuptools import setup, find_packages, Extension
# Import setupext IFF you use prep_cmd or want custom triggers
import setupext


os.chdir(os.path.dirname(sys.argv[0]) or ".")

'''
==============================================================================
PACKAGE DATA
==============================================================================
'''
# You _SHOULD_ set these
name = 'py25519'
version = '0.13.20'   # oldver: '0.13.9'
url = 'https://github.com/sundarnagarajan/py25519'
download_url = '%s/tree/%s' % (url, version)
description = name
install_requires = [
    'cffi_utils',
]
packages = find_packages()
license = 'License :: OSI Approved :: MIT License'

# The following are optional
author = 'Sundar Nagarajan'
# author_email = ''
maintainer = author
# maintainer_email = author_email
classifiers = [
    'Development Status :: 4 - Beta',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: Implementation :: PyPy',
    'License :: OSI Approved :: MIT License',
    ('License :: OSI Approved :: '
     'GNU Lesser General Public License v3 or later (LGPLv3+)'
     )
]
zip_safe = True


'''
==============================================================================
C EXTENSION DETAILS
C source files are NOT under the python module, so that the
C files are NOT installed with the python module

Add c_dir to MANIFEST.in as graft
Add any dir with prep scripts to MANIFEST.in as graft
==============================================================================
'''
c_dir = 'c'
libname = 'libcurve25519'
c_src_files = [
    'source/curve25519_dh.c',
    'source/curve25519_mehdi.c',
    'source/curve25519_order.c',
    'source/curve25519_utils.c',
    'source/custom_blind.c',
    'source/ed25519_sign.c',
    'source/ed25519_verify.c',
    'source/sha512.c',
]
libpath = os.path.join(name, libname)
c_src_list = [os.path.join(c_dir, x) for x in c_src_files]

# ext_modules should be a LIST of dict - each dict is a
# set of keywords that define ONE extension. For a SINGLE extension
# ext_modules should be a LIST with a SINGLE dict
ext_modules = [
    dict(
        name=libpath,
        sources=c_src_list,
        include_dirs=[c_dir],
    )
]


'''
==============================================================================
ADDITIONAL DATA FILES
==============================================================================
'''

data_dirs = [
    'doc',
]


'''
==============================================================================
CUSTOM STEPS
To just execute a (shell) script to prepare C sources, call
set prep_cmd = STRING containing command with all args
Typically you will pass c_dir as a single argument to allow script to
cd c_dir and then execute prep steps

prepare_c_source will automatically set setupext.config and cmdclass
==============================================================================
'''
prep_cmd = os.path.join('prep/build.sh') + ' ' + c_dir

'''
==============================================================================
ADDITIONAL keyword args to setup() - shouldn't be required, normally
==============================================================================
'''
ADDL_KWARGS = dict(
)


'''
==============================================================================
           DO NOT CHANGE ANYTHING BELOW THIS
==============================================================================
'''


def prepare_c_source(cmd):
    '''
    cmd-->str: command with arguments
    '''
    setupext.config['build_ext']['pre']['cmdlist'] = [cmd]
    return setupext.get_cmdclass()


def get_longdesc(default=''):
    '''
    Returns-->str
    '''
    files = ['README.rst', 'README.md', 'README.txt', 'README']
    for f in files:
        try:
            return open(f, 'r').read()
        except:
            continue
    return default


def get_dirtree(topdir, dirlist=[]):
    '''
    topdir-->str: must be name of a dir under current working dir
    dirlist-->list of str: must all be names of dirs under topdir
    '''
    ret = []
    curdir = os.getcwd()
    if not os.path.isdir(topdir):
        return ret
    os.chdir(topdir)
    try:
        for dirname in dirlist:
            if not os.path.isdir(dirname):
                continue
            for (d, ds, fs) in os.walk(dirname):
                for f in fs:
                    ret += [os.path.join(d, f)]
        return ret
    except:
        return ret
    finally:
        os.chdir(curdir)

# Make some keywords MANDATORY
for k in [
    'name', 'version', 'description', 'license',
]:
    if k not in locals():
        raise Exception('Missing mandatory keyword: ' + k)

# keywords that are computed from variables
dirlist = locals().get('data_dirs', None)
if isinstance(dirlist, list):
    package_dir = {name: name}
    package_data = {name: get_dirtree(topdir=name, dirlist=dirlist)}

long_description = get_longdesc(description)

known_keywords = [
    'name', 'version', 'packages', 'description', 'license',
    'install_requires', 'requires', 'setup_requires',
    'package_dir', 'package_data',
    'zip_safe', 'classifiers', 'keywords',
    'long_description', 'url', 'download_url',
    'author', 'author_email', 'maintainer', 'maintainer_email',
]

kwdict = {}
for k in known_keywords:
    if k in locals():
        kwdict[k] = locals()[k]

if 'prep_cmd' in locals():
    kwdict['cmdclass'] = prepare_c_source(locals()['prep_cmd'])

# Do not compile ext_modules during build phase - wasteful
if len(sys.argv) > 1 and sys.argv[1] != 'build':
    if 'ext_modules' in locals():
        kwdict['ext_modules'] = [Extension(**x) for x in
                                 locals()['ext_modules']]

# Additional keywords specified by user - shouldn't be required, normally
kwdict.update(ADDL_KWARGS)
setup(**kwdict)
